VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "CryptFile"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit
'=========
'CryptFile
'=========
'
'Encrypted file I/O object implemented in VB6.
'
'This is a Class written in VB6 for doing encrypted sequential binary and
'text I/O.  Encryption/decryption is done using the RC4 Class.
'
'Dependent on RC4 Class.
'
'Note that this Class was not intended to be compatible with any other
'software using the RC4 algorithm.  If you want that level of compatibility
'you may need to set the RC4 object's FMSSkip property to a non-default
'value (0 for straight RC4) and modify the KeyGen() function below.
'
'Properties
'----------
'
'   EOF As Boolean {Read Only}
'
'       o For write access it always returns True.
'
'       o For cfaReadBinary access it returns False until after the last
'         data has been returned by a ReadBytes() call.  If the last call
'         asked for more data than was left in the file, the available
'         bytes are returned and EOF becomes True.
'
'         This works differently than VB6's Get # does, and more like
'         text I/O (Input # or Line Input #) does.
'
'       o For cfaReadText access it returns False until after the last
'         line has been returned by a ReadLine() call.
'
'   IsOpen As Boolean {Read Only}
'
'       Returns True when CryptFile is open.
'
'   Loc As Long {Read Only}
'
'       Valid only for cfaReadBinary access.  Returns the same value as the
'       VB6 Loc() function.
'
'   LOF As Long {Read only}
'
'       Valid only for cfaReadBinary access.  Returns the same value as the
'       VB6 LOF() function.
'
'Methods
'-------
'
'   CloseFile()
'
'       Closes the file and disposes of buffers.  The CryptFile object may
'       then be used (opened) again.
'
'   CryptCopy(ByVal SrcFileName As String,
'             ByVal DestFileName As String,
'             ByVal Passphrase As String)
'
'       Copies any file applying RC4 (i.e. if the source file is encrypted
'       it creates a plaintext copy and vice versa).
'
'       If the destination file already exists it is deleted before copying
'       begins.
'
'       Passphrase must be at least one character, and if longer than 256
'       it is truncated to 256 characters.  This is used to create an RC4
'       Key of 1 to 256 bytes.  If characters outside the range 0 to 127
'       are used you may run into locale-related issues sharing encrypted
'       data files.
'
'   OpenFile(ByVal FileName As String,
'            ByVal Passphrase As String,
'            ByVal Access As cfAccesses,
'            Optional ByVal LfNewLine As Boolean = False,
'            Optional ByVal Unicode As Boolean = False,
'            Optional ByVal WriteBOM As Boolean = False)
'
'       Opens the file FileName using Passphrase for Access.
'
'       Passphrase is as described for CryptCopy() above.
'
'       NOTE:   When write access is requested any existing file is deleted
'               before opening the new empty file for writing!
'
'       When Access = cfaReadText:
'
'               LFNewLine = False means CrLf delimits lines.
'                         = True means Lf delimits lines.
'
'               Unicode   = False means ANSI to Unicode translation is done.
'                         = True means Unicode text is read without
'                           translation, and any UTF-16LE BOM found at the
'                           start of the file is skipped.
'
'       When Access = cfaWriteText:
'
'               LFNewLine = False means CrLf delimits lines.
'                         = True means Lf delimits lines.
'
'               Unicode   = False means Unicode to ANSI translation is done.
'                         = True means Unicode text is written without
'                           translation, and if WriteBOM is True a UTF-16LE
'                           BOM is written at the start of the file.
'
'       When Access = cfaReadBinary or = cfaWriteBinary:
'
'               LFNewLine, Unicode, and WriteBOM are ignored.
'
'   ReadBytes(ByRef Bytes() As Byte)
'
'       Valid only for cfaReadBinary access.  Reads/fills Bytes() array with
'       decrytped data.  Bytes() must be pre-allocated to the size to be
'       read, much as when using VB6's Get # statement.
'
'   ReadLine() As String
'
'       Valid only for cfaReadText access.  Returns a decrypted line of
'       text.
'
'   WriteBytes(ByRef Bytes() As Byte)
'
'       Valid only for cfaWriteBinary access.  Encrypts and writes Bytes()
'       to the file.  Bytes() is passed ByRef but will not be altered.
'
'   WriteLine(ByRef Line As String)
'
'       Valid only for cfaWriteText access.  Encrypts and writes Line to
'       the file.  Line is passed ByRef but will not be altered.
'

Public Enum cfAccesses
    cfaReadText
    cfaWriteText
    cfaReadBinary
    cfaWriteBinary
End Enum
'Hack: Preserve case of these Enum value names when typed in calling code:
#If False Then
    Dim cfaReadText, cfaWriteText, cfaReadBinary, cfaWriteBinary
#End If

'BLOCKSIZE only applies to text access and CryptCopy, and is not used for
'binary reading or writing.  It must be a multiple of 2 in order to
'gracefully handle Unicode.  Optimal values normally range from 16384 to
'65536.
Private Const BLOCKSIZE As Long = 32768

'FMSSKIP_VALUE is normally set at -1 in order to use the RC4 object's
'default value.  See the RC4 Class for more details on alternative values.
'Normally you'd only chage this for compatibility with other encryption
'software.
Private Const FMSSKIP_VALUE As Long = -1

Private RC4 As RC4
Private cfaAccess As cfAccesses
Private blnUnicode As Boolean
Private blnLfNewLine As Boolean
Private strNewLine As String
Private strCr As String
Private intFile As Integer
Private lngBytesLeft As Long
Private bytBlock() As Byte
Private strBlock As String
Private lngBlockSize As Long
Private lngBlockSizeU As Long
Private lngNext As Long

Public Property Get EOF() As Boolean
    Dim lngBlock As Long
    
    If Not IsOpen Then Err.Raise 52 'Bad file name or number.
    
    Select Case cfaAccess
        Case cfaReadText
            If blnUnicode Then
                lngBlock = lngBlockSizeU
            Else
                lngBlock = lngBlockSize
            End If
            EOF = (lngBytesLeft < 1) And (lngNext > lngBlock)
            
        Case cfaReadBinary
            EOF = lngBytesLeft < 1
            
        Case Else
            EOF = True
    End Select
End Property

Public Property Get IsOpen() As Boolean
    IsOpen = intFile <> 0
End Property

Public Property Get Loc() As Long
    If Not IsOpen Then Err.Raise 52 'Bad file name or number.
    If cfaAccess <> cfaReadBinary And cfaAccess <> cfaWriteBinary Then Err.Raise 54 'Bad file mode.
    
    Loc = VBA.Loc(intFile)
End Property

Public Property Get LOF() As Long
    If Not IsOpen Then Err.Raise 52 'Bad file name or number.
    If cfaAccess <> cfaReadBinary And cfaAccess <> cfaWriteBinary Then Err.Raise 54 'Bad file mode.
    
    LOF = VBA.LOF(intFile)
End Property

Public Sub CloseFile()
    If IsOpen Then
        Select Case cfaAccess
            Case cfaReadText
                Erase bytBlock
                
            Case cfaWriteText
                If lngNext > 1 Then PutBlock
                strBlock = ""
        End Select
        Close #intFile
        intFile = 0
        Set RC4 = Nothing
    End If
End Sub

Public Sub CryptCopy( _
    ByVal SrcFileName As String, _
    ByVal DestFileName As String, _
    ByVal Passphrase As String)
    
    Dim intSrcFile As Integer
    Dim intDestFile As Integer
    
    If Len(Passphrase) < 1 Then Err.Raise 5 'Invalid procedure call.
    On Error Resume Next
    GetAttr SrcFileName
    If Err Then Err.Clear: Err.Raise 53 'File not found.
    Kill DestFileName
    On Error GoTo 0
    
    Set RC4 = New RC4
    RC4.FMSSkip = FMSSKIP_VALUE
    RC4.SetSecret KeyGen(Passphrase)
    intSrcFile = FreeFile(1)
    intDestFile = FreeFile(0)
    Open SrcFileName For Binary Access Read As #intSrcFile
    lngBytesLeft = VBA.LOF(intSrcFile)
    lngBlockSize = BLOCKSIZE
    ReDim bytBlock(1 To lngBlockSize)
    Open DestFileName For Binary Access Write As #intDestFile
    
    Do While lngBytesLeft > 0
        If lngBytesLeft < lngBlockSize Then
            lngBlockSize = lngBytesLeft
            ReDim bytBlock(1 To lngBlockSize)
        End If
        Get #intSrcFile, , bytBlock
        lngBytesLeft = lngBytesLeft - lngBlockSize
        RC4.CryptBlock bytBlock
        Put #intDestFile, , bytBlock
    Loop
    
    Close #intDestFile, #intSrcFile
End Sub

Public Sub OpenFile( _
    ByVal FileName As String, _
    ByVal Passphrase As String, _
    ByVal Access As cfAccesses, _
    Optional ByVal LfNewLine As Boolean = False, _
    Optional ByVal Unicode As Boolean = False, _
    Optional ByVal WriteBOM As Boolean = False)
    
    If Len(Passphrase) < 1 Then Err.Raise 5 'Invalid procedure call.
    If IsOpen Then Err.Raise 55 'File already open.
    
    cfaAccess = Access
    If cfaAccess = cfaWriteText Or cfaAccess = cfaWriteBinary Then
        On Error Resume Next
        Kill FileName
        On Error GoTo 0
    Else
        On Error Resume Next
        GetAttr FileName
        If Err Then Err.Clear: Err.Raise 53 'File not found.
        On Error GoTo 0
    End If
    blnUnicode = Unicode
    blnLfNewLine = LfNewLine
    Set RC4 = New RC4
    RC4.FMSSkip = FMSSKIP_VALUE
    RC4.SetSecret KeyGen(Passphrase)
    Select Case cfaAccess
        Case cfaReadText
            intFile = FreeFile(1)
            Open FileName For Binary Access Read As #intFile
            lngBytesLeft = VBA.LOF(intFile)
            If lngBytesLeft > 0 Then
                SetNewLine
                lngBlockSize = BLOCKSIZE
                ReDim bytBlock(1 To lngBlockSize)
                If blnUnicode Then lngBlockSizeU = lngBlockSize \ 2&
                GetBlock
                If blnUnicode Then
                    If bytBlock(1) = &HFF And bytBlock(2) = &HFE Then lngNext = 2
                End If
            End If
            
        Case cfaWriteText
            intFile = FreeFile(0)
            Open FileName For Binary Access Write As #intFile
            SetNewLine
            lngBlockSize = BLOCKSIZE
            strBlock = Space$(lngBlockSize)
            lngNext = 1
            If blnUnicode Then
                lngBlockSizeU = lngBlockSize \ 2&
                If WriteBOM Then
                    Mid$(strBlock, 1, 1) = ChrW$(&HFEFF)
                    lngNext = 2
                End If
            End If
        
        Case cfaReadBinary
            intFile = FreeFile(1)
            Open FileName For Binary Access Read As #intFile
            lngBytesLeft = VBA.LOF(intFile)
        
        Case cfaWriteBinary
            intFile = FreeFile(0)
            Open FileName For Binary Access Write As #intFile
        
        Case Else
            Err.Raise 5 'Invalid procedure call.
    End Select
End Sub

Public Sub ReadBytes(ByRef Bytes() As Byte)
    Dim lngBytesSize As Long
    
    If Not IsOpen Then Err.Raise 52 'Bad file name or number.
    If cfaAccess <> cfaReadBinary Then Err.Raise 54 'Bad file mode.
    If lngBytesLeft < 1 Then Err.Raise 62 'Input past end of file.
    
    lngBytesSize = (UBound(Bytes) - LBound(Bytes) + 1)
    Get #intFile, , Bytes
    RC4.CryptBlock Bytes
    If lngBytesSize > lngBytesLeft Then Err.Raise 62 'Input past end of file (but read some data).
    lngBytesLeft = lngBytesLeft - lngBytesSize
End Sub

Public Function ReadLine() As String
    Dim lngEnd As Long
    
    If Not IsOpen Then Err.Raise 52 'Bad file name or number.
    If cfaAccess <> cfaReadText Then Err.Raise 54 'Bad file mode.
    If EOF Then Err.Raise 62 'Input past end of file.

    Do
        If blnUnicode Then
            lngEnd = InStr(lngNext, bytBlock, strNewLine)
        Else
            lngEnd = InStrB(lngNext, bytBlock, strNewLine)
        End If
        If lngEnd > 0 Then
            'Extract the rest of the line.
            If blnUnicode Then
                ReadLine = ReadLine & Mid$(bytBlock, lngNext, lngEnd - lngNext)
            Else
                ReadLine = ReadLine _
                         & StrConv(MidB$(bytBlock, lngNext, lngEnd - lngNext), vbUnicode)
            End If
            lngNext = lngEnd + IIf(blnLfNewLine, 1, 2)
            If blnUnicode Then
                If lngNext <= lngBlockSizeU Then Exit Do
            Else
                If lngNext <= lngBlockSize Then Exit Do
            End If
            GetBlock
            Exit Do
        Else
            If Not blnLfNewLine Then
                If blnUnicode Then
                    lngEnd = InStr(lngNext, bytBlock, strCr)
                Else
                    lngEnd = InStrB(lngNext, bytBlock, strCr)
                End If
                If lngEnd > 0 Then
                    'Extract the rest of the line.
                    If blnUnicode Then
                        ReadLine = ReadLine & Mid$(bytBlock, lngNext, lngEnd - lngNext)
                    Else
                        ReadLine = ReadLine _
                                 & StrConv(MidB$(bytBlock, lngNext, lngEnd - lngNext), vbUnicode)
                    End If
                    GetBlock
                    lngNext = 2 'Skip Lf.
                    Exit Do
                End If
            End If
            'Didn't find end, extract the part we have.
            If blnUnicode Then
                ReadLine = ReadLine & Mid$(bytBlock, lngNext)
                lngNext = lngBlockSizeU + 1
            Else
                ReadLine = ReadLine & StrConv(MidB$(bytBlock, lngNext), vbUnicode)
                lngNext = lngBlockSize + 1
            End If
            GetBlock
        End If
    Loop Until EOF
End Function

Public Sub WriteBytes(ByRef Bytes() As Byte)
    If Not IsOpen Then Err.Raise 52 'Bad file name or number.
    If cfaAccess <> cfaWriteBinary Then Err.Raise 54 'Bad file mode.
            
    bytBlock = Bytes
    RC4.CryptBlock bytBlock
    Put #intFile, , bytBlock
    Erase bytBlock
End Sub

Public Sub WriteLine(ByRef Line As String)
    Dim lngLen As Long
    Dim lngNLLen As Long
    
    If Not IsOpen Then Err.Raise 52 'Bad file name or number.
    If cfaAccess <> cfaWriteText Then Err.Raise 54 'Bad file mode.
        
    lngLen = Len(Line)
    lngNLLen = IIf(blnLfNewLine, 1, 2)
    If blnUnicode Then
        If lngNext + lngLen + lngNLLen > lngBlockSizeU Then
            'Line plus newline doesn't fit in remaining block space.
            'Write the current block if it has any data and then write
            'this line (as a block on its own, because it might exceed
            'BLOCKSIZE).
            If lngNext > 1 Then PutBlock
            strBlock = Line
            lngNext = lngLen + 1
            PutBlock
            lngLen = 0
            strBlock = Space$(lngBlockSize)
        Else
            'Fits, add to block.
            Mid$(strBlock, lngNext, lngLen) = Line
        End If
        Mid$(strBlock, lngNext + lngLen, lngNLLen) = strNewLine
    Else
        If lngNext + lngLen + lngNLLen > lngBlockSize Then
            If lngNext > 1 Then PutBlock
            strBlock = StrConv(Line, vbFromUnicode)
            lngNext = lngLen + 1
            PutBlock
            lngLen = 0
            strBlock = Space$(lngBlockSize)
        Else
            MidB$(strBlock, lngNext, lngLen) = StrConv(Line, vbFromUnicode)
        End If
        MidB$(strBlock, lngNext + lngLen, lngNLLen) = strNewLine
    End If
    lngNext = lngNext + lngLen + lngNLLen
End Sub

Private Function KeyGen(ByRef Passphrase As String) As Byte()
    Dim bytKey() As Byte
    Dim I As Long
    
    bytKey = StrConv(Left$(Passphrase, 256), vbFromUnicode)
    For I = UBound(bytKey) Mod 2 To UBound(bytKey) Step 2
        bytKey(I) = bytKey(I) Xor &H80
    Next
    KeyGen = bytKey
End Function

Private Sub GetBlock()
    If lngBytesLeft > 0 Then
        If lngBytesLeft < lngBlockSize Then
            lngBlockSize = lngBytesLeft
            ReDim bytBlock(1 To lngBlockSize)
            If blnUnicode Then lngBlockSizeU = lngBlockSize \ 2&
        End If
        Get #intFile, , bytBlock
        lngBytesLeft = lngBytesLeft - lngBlockSize
        RC4.CryptBlock bytBlock
        lngNext = 1
    End If
End Sub

Private Sub PutBlock()
    If blnUnicode Then
        bytBlock = Left$(strBlock, lngNext - 1)
    Else
        bytBlock = LeftB$(strBlock, lngNext - 1)
    End If
    RC4.CryptBlock bytBlock
    Put #intFile, , bytBlock
    Erase bytBlock
    lngNext = 1
End Sub

Private Sub SetNewLine()
    If blnUnicode Then
        If blnLfNewLine Then
            strNewLine = vbLf
        Else
            strNewLine = vbNewLine
            strCr = vbCr
        End If
    Else
        If blnLfNewLine Then
            strNewLine = StrConv(vbLf, vbFromUnicode)
        Else
            strNewLine = StrConv(vbNewLine, vbFromUnicode)
            strCr = StrConv(vbCr, vbFromUnicode)
        End If
    End If
End Sub

Private Sub Class_Terminate()
    If IsOpen Then CloseFile
End Sub
