VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "RC4"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit
'===
'RC4
'===
'
'RC4/RC4-drop(n) stream cipher implemented in VB6.
'
'This is a moderate performance implementation of the "Alleged RC4" symmetric
'stream cipher.  The relative simplicity of RC4 makes even this VB6
'implementation's performance more than adequate in many cases.
'
'Here we discard the first 3000 bytes of the stream by default to aid in
'defeating Fluhrer, Martin and Shamir Attacks.  This can be changed via the
'FMSSkip property.  This modification to RC4 is often called RC4-drop(n).
'
'However FMS and other cracks such as Klein's Attack are only effective when
'there are tens of thousands of samples to analyze that reuse the same key.
'While this is a problem for an encrypted communication channel (SSL, WEP) it
'is much less so for encrypting files.
'
'So despite the known cracks RC4 is still useful for many applications,
'especially when the -drop(n) enhancement is applied.
'
'
'Using RC4:
'
'   o Create an instance of the RC4 Class.
'   o Set a value for FMSSkip if the default is not desired.
'   o Call SetSecret() passing a binary Key to initialize the Class.
'   o Call CryptBlock() for each block of Content until the stream has been
'     completely processed.
'
'   An instance of RC4 can be reused for another stream by calling SetSecret()
'   again.
'
'Properties
'----------
'
'   FMSSkip As Long {R/W}
'
'       Must be set prior to calling SetSecret() in order to be effective!
'
'       Defaults to -1 (meaning "use 3000").  Can be set to -1, or a values
'       from 0 to an appropriate value as needed for compatibility with other
'       RC4 implementations. Maximum value accepted is 8192.
'
'       Setting FMSSkip = 0 disables FMS remediation.
'
'       Chosing "weird" values besides 0 or 3000 might be considered a way of
'       adding further effective Key randomization but this isn't generally
'       recommended.  If you do this "good" values are considered to be those
'       in the range from 768 to 3072
'
'Methods
'-------
'
'   CryptBlock(ByRef Content() As Byte)
'
'       Once Secret has been set you can make repeated calls to CryptBlock(),
'       passing a block of binary content in each call.
'
'       CryptBlock accepts a Byte array with any bounds, and returns with it
'       changed: encrypted or decrypted.
'
'       This is the default method of the Class.
'
'   SetSecret(ByRef Key() As Byte)
'
'       Call SetSecret() passing a binary key to initialize the Class for
'       processing a new stream.
'
'       Secret accepts a Byte array key with a lower bound of 0 and no more
'       than 256 bytes of keydata.
'
'       Key() is not altered by this process.
'

Private S(255) As Integer
Private I As Long
Private J As Long
Private lngFMSSkip As Long

Public Property Get FMSSkip() As Long
    FMSSkip = lngFMSSkip
End Property

Public Property Let FMSSkip(ByVal FMSSkip As Long)
    If -1 > FMSSkip Or FMSSkip > 8192 Then Err.Raise 5 'Invalid procedure call.
    lngFMSSkip = FMSSkip
End Property

Public Sub SetSecret(ByRef Key() As Byte)
    Dim KeyLength As Long
    Dim TempSsubI As Long
    Dim Defeat_FMS_Attacks() As Byte
    
    If LBound(Key) <> 0 Or UBound(Key) > 255 Then
        Err.Raise 5 'Invalid procedure call.
    Else
        KeyLength = UBound(Key) + 1
        For I = 0 To 255
            S(I) = I
        Next
        J = 0
        For I = 0 To 255
            TempSsubI = S(I)
            J = (J + TempSsubI + Key(I Mod KeyLength)) Mod 256
            S(I) = S(J)
            S(J) = TempSsubI
        Next
    End If
    
    I = 0
    J = 0
    
    If lngFMSSkip = -1 Then
        ReDim Defeat_FMS_Attacks(2999)
    ElseIf lngFMSSkip > 0 Then
        ReDim Defeat_FMS_Attacks(lngFMSSkip - 1)
    Else
        Exit Sub
    End If
    CryptBlock Defeat_FMS_Attacks
End Sub

Public Sub CryptBlock(ByRef Content() As Byte)
Attribute CryptBlock.VB_UserMemId = 0
    'Default method.
    Dim N As Long
    Dim TempSsubI As Long
    
    For N = LBound(Content) To UBound(Content)
        I = (I + 1) Mod 256
        TempSsubI = S(I)
        J = (J + TempSsubI) Mod 256
        S(I) = S(J)
        S(J) = TempSsubI
        Content(N) = Content(N) Xor (S((S(I) + TempSsubI) Mod 256))
    Next
End Sub

Private Sub Class_Initialize()
    lngFMSSkip = -1 'Default value.
End Sub
